#define LCD_ANZEIGE

// Anschlu einer LC-Anzeige 2x16 
// lcd_init() muss als erste Funktion aufgerufen werden. Danach sollten
// lcd_cmd() und lcd_zeichen() Befehle und Daten korrekt ausgeben.

/*
Das LCD-Modul wird im 4-Bit Modus betrieben. D0-D3 liegen daher auf GND.
Verwendet werden von PortD die Pins 9-15 wie folgt:

  LCD-Pin   Funktion    Discovery-Board

    1         GND            GND-Pin
    2         +5V        +5V ber NC-Pin
    3       Kontrast    per DAC_1 oder Diode->GND (ca.0,7V Vorsp.)
    4         RS              PD9
    5         R/W             PD10
    6         E               PD11
    7       D0(GND)
    8       D1(GND)
    9       D2(GND)
    10      D3(GND)
    11        D4              PD12
    12        D5              PD13
    13        D6              PD14
    14        D7              PD15

2020-10-24
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/


#include "stm32f4xx.h"

#define BIT(x) (1<<x)
enum gpio_modus {IN, OUT, AF, ANALOG};
enum gpio_speed {LOW_SPEED, MEDIUM_SPEED, HIGH_SPEED, V_HIGH_SPEED};
enum gpio_pullup {NO_PULL, PULL_UP, PULL_DOWN};

#define CLR_MODER         3         // Lschmaske
#define SET_GPIO          0x1       // untere Bits
#define RESET_GPIO        0x10000   // obere Bits

uint8_t lcd_vorhanden=1;            // default

#define DEF_LCD_KONTRAST  20        // DAC-Wert; sinnvoll: 20 - 50
#define DEF_LCD_BREITE    16        

#define LCD_CMD     9               // 0 = CMD, 1 = Zeichenausgabe
#define LCD_READ    10              // 0 = Schreiben, 1 = Status lesen
#define LCD_STROBE  11              // Schreibimpuls E fuers LCD
#define LCD_D4      12              // Datenleitungen
#define LCD_D5      13
#define LCD_D6      14
#define LCD_D7      15
#define ZEILE1      0x7f            // jeweils 1 addieren fuer 1. Spalte
#define ZEILE2      0xbf            // dto.
#define BUSY_FLAG   0x80            // Bit7 vom LCD-Status

uint8_t zeilen_tab[] = {0,ZEILE1,ZEILE2,ZEILE1,ZEILE2}; /* alle werte -1 */
uint8_t lcd_zeile_start[5];
uint8_t lcd_aktiv;
uint8_t lcd_breite;
uint8_t lcd_kontrast = DEF_LCD_KONTRAST;

void lcd_impuls();

// direktes Setzen oder Ruecksetzen eines Port-Pins
static void set_pin(GPIO_TypeDef* GPIOx, uint8_t number, char set) {
  if(set) {
    GPIOx->BSRR = (SET_GPIO << number);       // IO-Bit setzen
  } else {
    GPIOx->BSRR = (RESET_GPIO << number);     // IO-Bit loeschen
  }
}

// 16 oder 20 Zeichen/Zeile initialisieren
void init_zeilen_breite(uint8_t breite)
{
int i, offset=1;
  if(breite != DEF_LCD_BREITE) offset += (breite-DEF_LCD_BREITE)/2;       // 2 Stellen weiter nach rechts schreiben
  lcd_breite = breite;
  zeilen_tab[3] = ZEILE1 + breite;            // fr Zeile 3
  zeilen_tab[4] = ZEILE2 + breite;            // fr Zeile 4
  for(i=0; i<5; i++)
    lcd_zeile_start[i] = zeilen_tab[i]+offset;
} 

// unkritisch: ungefhrer Wert 1,3 s
void warte_1us(void)
{
static uint32_t init; 
  if(!init) {
    init = 1;
    RCC->APB1ENR |= RCC_APB1ENR_TIM13EN;      // Timer13 aktivieren
    TIM13->CR1 |= TIM_CR1_CEN;                // T13 starten
  }  
  TIM13->CNT = -(SystemCoreClock/1000000); 
  TIM13->SR &= 0xfffe;          // berlauf flag lschen 
  while(!(TIM13->SR & 1));      // und auf berlauf warten
}

// LCD kontrast per DAC1
void setze_kontrast(uint8_t wert)
{
static uint8_t init;
  if(!init) {
    init = 1;
    RCC->APB1ENR |= RCC_APB1ENR_DACEN;
    DAC->CR = DAC_CR_EN1;
  }
  DAC->DHR8R1 = wert;
}  
  
uint8_t lcd_lesen(void)                // oberstes Bit ist busy-flag, Rest: Adresse
{
  uint8_t temp = 0;
// Datenleitungen auf Eingang
  GPIOD->MODER &= ~(CLR_MODER<<LCD_D4*2 | 
                    CLR_MODER<<LCD_D5*2 | 
                    CLR_MODER<<LCD_D6*2 | 
                    CLR_MODER<<LCD_D7*2);
  
  set_pin(GPIOD, LCD_READ, 1);          // Lesen aktivieren
  warte_1us();
  set_pin(GPIOD, LCD_STROBE, 1);        // E zum Lesen setzen
  warte_1us();
  temp = (GPIOD->IDR >> 8) & 0xf0;      // oberes nibble zuerst
  set_pin(GPIOD, LCD_STROBE, 0);        // und wieder auf 0
  warte_1us();
  set_pin(GPIOD, LCD_STROBE, 1);        // nchstes nibble lesen
  warte_1us();
  temp |= (GPIOD->IDR >> 12) & 0x0f;    // unteres nibble zuletzt
  set_pin(GPIOD, LCD_STROBE, 0);        // und wieder auf 0
  set_pin(GPIOD, LCD_READ, 0);          // dto
  warte_1us();
  
// Datenleitungen wieder auf Ausgang  
  GPIOD->MODER |= (OUT<<LCD_D4*2 | 
                   OUT<<LCD_D5*2 | 
                   OUT<<LCD_D6*2 | 
                   OUT<<LCD_D7*2);
  warte_1us();
  return(temp);                         // busy-flag + Adresse
}

// Status lesen; busy-flag == 0, wenn neue Ausgabe mglich
uint8_t lcd_status(void)                // oberstes Bit ist busy-flag, Rest: Adresse
{
  set_pin(GPIOD, LCD_CMD, 0);           // CMD ausfhren
  warte_1us();
  return(lcd_lesen());
} 


void lcd_impuls(void)                   // Schreibimpuls mit Mindestlaenge
{
  warte_1us();
  set_pin(GPIOD, LCD_STROBE, 1);        // 1 = aktiv
  warte_1us();
  set_pin(GPIOD, LCD_STROBE, 0);        // und wieder auf 0
  warte_1us();
}

void lcd_nibble(int8_t c)               // 4-bit-Wert an PD12-PD15 ausgeben
{
  GPIOD->BSRR = (((c & 0xf0) << 8) ^ 0xf000 << 16);   // alle Datenleitungen auf 0
  GPIOD->BSRR = (c & 0xf0) << 8;        // danach aktive Bits setzen
  lcd_impuls();
}

void lcd_out(int8_t z, int8_t mode)     // Ausgabe von Daten (mode==1) oder CMDs (mode==0)
{
    while(lcd_status() & BUSY_FLAG);    // vor neuer Ausgabe busy-flag abwarten
    if(mode) set_pin(GPIOD, LCD_CMD, 1);
    else set_pin(GPIOD, LCD_CMD, 0);
    lcd_nibble(z);
    lcd_nibble(z<<4);
}

uint8_t lcd_test(void)               
{
uint8_t temp = 0;
  set_pin(GPIOD, LCD_CMD, 1);  // immer setzen, um Daten zu lesen
  warte_1us();
  temp = lcd_lesen();
  return(temp);
}  

void lcd_cmd(uint8_t z)        // einen LCD-Befehl ausgeben
{
  lcd_out(z,0);
}

void lcd_zeichen(char z)    // ein Zeichen ausgeben
{
  lcd_out(z,1);
}

void lcd_string(char *s)    // Zeichenkette auf LCD
{
int8_t temp;
  if(lcd_vorhanden) {
   while((temp=*s++) != 0) {
     lcd_zeichen(temp);
   }
  }
}

void lcd_init_nibble(int8_t c)  // 4-bit-Wert ausgeben LCD_INIT
{
uint16_t temp = 10000;          // 10 ms
  lcd_nibble(c);
  while(temp--) warte_1us();    // bei Init genug Zeit lassen
}

void lcd_zeile1(char *s)      // komplette Zeile ausgeben
{
  if(lcd_vorhanden) {
    lcd_cmd(lcd_zeile_start[1]);// 1.Zeile, 1.Spalte
    lcd_string(s);
  }
}
void lcd_zeile2(char *s)      // komplette Zeile ausgeben
{
  if(lcd_vorhanden) {
    lcd_cmd(lcd_zeile_start[2]);// 2.Zeile, 1.Spalte
    lcd_string(s);
  }
}


void lcd_cursor_on()
{
  lcd_cmd(0xf);
}

void lcd_cursor_off()
{
  lcd_cmd(0xc);
}

void lcd_zeile_spalte(uint8_t zeile, uint8_t spalte)  // Schreibposition setzen
{
  if(zeile && spalte) lcd_cmd(zeilen_tab[zeile] + spalte);
}

void loesche_lcd_zeile(uint8_t zeile, uint8_t n)
{
uint8_t i;  
  lcd_cmd(zeilen_tab[zeile] + 1);         // Startposition
  for(i=0; i<n; i++) lcd_zeichen(' ');    // Leerzeichen schreiben
}  

void lcd_clear()
{
  lcd_cursor_off();
  lcd_cmd(1);
}

void loesche_lcd_zeile1()
{
  loesche_lcd_zeile(1, lcd_breite);
}
void loesche_lcd_zeile2()
{
  loesche_lcd_zeile(2, lcd_breite);
}
void loesche_lcd_zeile12()
{
  loesche_lcd_zeile(1, lcd_breite);
  loesche_lcd_zeile(2, lcd_breite);
}

  
uint8_t lcd_init(void)                    // Port D9 - PortD 15 werden gebraucht
{
uint32_t temp;
  RCC->AHB1ENR |= (RCC_AHB1ENR_GPIODEN);  // PortD aktivieren

// Ausgnge von PortD Pins 9, 10, 11, 12, 13, 14, 15 push-pull output*/
  GPIOD->MODER |= (OUT << LCD_CMD*2 | 
                   OUT << LCD_READ*2 | 
                   OUT << LCD_STROBE*2 | 
                   OUT << LCD_D4*2 | 
                   OUT << LCD_D5*2 | 
                   OUT << LCD_D6*2 | 
                   OUT << LCD_D7*2);
  GPIOD->PUPDR |= (PULL_DOWN << LCD_D4*2 |  // fr Statusabfrage
                   PULL_DOWN << LCD_D5*2 | 
                   PULL_DOWN << LCD_D6*2 | 
                   PULL_DOWN << LCD_D7*2);  // ohne LCD liefert busy-Abfrage "bereit"
  
  init_zeilen_breite(DEF_LCD_BREITE);
  temp = 50000;                           // 50 ms
  while(temp--) warte_1us();              // nach Reset genug Zeit lassen

  lcd_init_nibble(0x30);      // 8 Bit Modus
  lcd_init_nibble(0x30);      // wiederholen
  lcd_init_nibble(0x30);      // wiederholen
  lcd_init_nibble(0x20);      // dann auf 4 Bit Modus
  lcd_cmd(0x28);              // 2 oder 4 Zeilen
  lcd_cmd(0x28);              // wiederholen

  lcd_cmd(0x8);               // LCD off            
  lcd_cmd(0x6);               // LCD auto. increment
  lcd_cmd(0x17);              // LCD char. mode; PWR on
  lcd_cmd(0x1);               // LCD clear
  lcd_cmd(0x2);               // LCD home
  
  lcd_cmd(0xc);               // LCD on; ohne Cusor und Blinken
  lcd_cmd(1);                 // LCD clear
  setze_kontrast(lcd_kontrast);

  temp = 30000;                           // 30 ms
  while(temp--) warte_1us();              // nach init genug Zeit lassen

  lcd_aktiv = 1;
  return(lcd_status());
}

